import math
import sympy as sym
from typing import Union, List, Dict, Tuple
import numpy as np

import renderer.internals.internal as internal # type: ignore
import renderer.tools as tools
import renderer.validate as validate
import renderer.misc as misc

RealNum = Union[int, float]
Coord = Tuple[RealNum, RealNum]
TileCoord = Tuple[int, int, int]

def midpoint(x1: RealNum, y1: RealNum, x2: RealNum, y2: RealNum, o: RealNum, n: int=1, returnBoth: bool=False) -> Union[List[Tuple[RealNum, RealNum, RealNum]], List[List[Tuple[RealNum, RealNum, RealNum]]]]:
    """
    Calculates the midpoint of two lines, offsets the distance away from the line, and calculates the rotation of the line.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.midpoint
    """
    #print(x1, y1, x2, y2, o, n)
    points = []
    for p in range(1, n+1):
        x3, y3 = (x1+p*(x2-x1)/(n+1),y1+p*(y2-y1)/(n+1))
        #print(x3, y3)
        #x3, y3 = ((x1+x2)/2, (y1+y2)/2)
        if x1 == x2:
            m1 = None
            m2 = 0
        elif y1 == y2:
            m1 = 0
            m2 = None
        else:
            m1 = (y2-y1)/(x2-x1)
            m2 = -1 / m1
        results = pointsAway(x3, y3, o, m2)
        if returnBoth:
            #print(eq1, eq2)
            rot = 90 if x1 == x2 else math.degrees(-math.atan(m1))
            try:
                points += [(results[0][0], results[0][1], rot), (results[1][0], results[1][1], rot)]
            except IndexError:
                pass
        #print(results)
        elif x1 == x2:
            if o < 0:
                x4, y4 = results[0] if results[0][0] < results[1][0] else results[1]
            else:
                x4, y4 = results[0] if results[0][0] > results[1][0] else results[1]
            rot = 90
            points.append((x4, y4, rot))
        else:
            if o < 0:
                x4, y4 = results[0] if results[0][1] < results[1][1] else results[1]
            else:
                x4, y4 = results[0] if results[0][1] > results[1][1] else results[1]
            rot = math.degrees(-math.atan(m1))
            points.append((x4, y4, rot))
    return points

def linesIntersect(x1: RealNum, y1: RealNum, x2: RealNum, y2: RealNum, x3: RealNum, y3: RealNum, x4: RealNum, y4: RealNum) -> bool:
    """
    Finds if two segments intersect.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.linesIntersect
    """
    xv, yv = sym.symbols('xv,yv')
    if x1 == x2:
        m1 = None
        eq1 = sym.Eq(xv,x1)
        c1 = None if x1 != 0 else math.inf
    else:
        m1 = (y2-y1)/(x2-x1)
        eq1 = sym.Eq(yv-y1,m1*(xv-x1))
        c1 = y1-m1*x1
    if x3 == x4:
        m2 = None
        eq2 = sym.Eq(xv,x3)
        c2 = None if x3 != 0 else math.inf
    else:
        m2 = (y4-y3)/(x4-x3)
        eq2 = sym.Eq(yv-y3,m2*(xv-x3))
        c2 = y3-m2*x3
    if m1 == m2 and c1 == c2: #same eq
        if x1 == x2:
            return False if (min(y1,y2)>max(y3,y4) and max(y1,x2)>min(y3,y4)) or (min(y3,y4)>max(y1,y2) and max(y3,y4)>min(y1,y2)) else True
        else:
            return False if (min(x1,x2)>max(x3,x4) and max(x1,x2)>min(x3,x4)) or (min(x3,x4)>max(x1,x2) and max(x3,x4)>min(x1,x2)) else True
    elif m1 == m2: #parallel
        return False
    #print(eq1, eq2)
    result = sym.solve([eq1, eq2], (xv, yv))
    if isinstance(result, list) and result != []:
        x5, y5 = result[0]
    elif isinstance(result, dict):
        x5 = result[xv]
        y5 = result[yv]
    else:
        return False
    x1 = round(x1, 10); x2 = round(x2, 10); x3 = round(x3, 10); x4 = round(x4, 10); x5 = round(x5, 10)
    y1 = round(y1, 10); y2 = round(y2, 10); y3 = round(y3, 10); y4 = round(y4, 10); y5 = round(y5, 10)
    return False if (x5>max(x1,x2) or x5<min(x1,x2) or y5>max(y1,y2) or y5<min(y1,y2) or x5>max(x3,x4) or x5<min(x3,x4) or y5>max(y3,y4) or y5<min(y3,y4)) else True

def pointInPoly(xp: RealNum, yp: RealNum, coords: List[Coord]) -> bool:
    """
    Finds if a point is in a polygon. WARNING: If your polygon has a lot of corners, this will take very long.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.pointInPoly
    """
    r = -math.inf
    for x, _ in coords:
        if x > r:
            r = x
    r += 10
    cross = 0
    for i in range(len(coords)):
        j = i + 1 if i != len(coords)-1 else 0
        if linesIntersect(coords[i][0], coords[i][1], coords[j][0], coords[j][1], xp, yp, r, yp):
            if coords[i][1] == yp:
                h = i - 1 if i != 0 else len(coords)-1
                cross += 0 if (coords[h][1]>yp and coords[i][1]>yp) or (coords[h][1]<yp and coords[i][1]<yp) else 0.5
            elif coords[j][1] == yp:
                k = j + 1 if j != len(coords)-1 else 0 
                cross += 0 if (coords[j][1]>yp and coords[k][1]>yp) or (coords[j][1]<yp and coords[k][1]<yp) else 0.5
            else:
                cross += 1
    return cross % 2 == 1
            
def polyCenter(coords: List[Coord]) -> Coord:
    """
    Finds the center point of a polygon.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.polyCenter
    """
    mx = []
    my = []
    for x1, y1 in coords:
        for x2, y2 in coords:
            x3, y3, _ = midpoint(x1, y1, x2, y2, 0)[0]
            mx.append(x3)
            my.append(y3)
    return sum(mx)/len(mx), sum(my)/len(my)

def lineInBox(line: List[Coord], top: RealNum, bottom: RealNum, left: RealNum, right: RealNum) -> bool:
    """
    Finds if any nodes of a line go within the box.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.lineInBox
    """
    for i in range(len(line)-1):
        x1, y1 = line[i]
        x2, y2 = line[i+1]
        for y3, x3, y4, x4 in [(top,left,bottom,left), (bottom,left,bottom,right), (bottom,right,top,right), (top,right,top,left)]:
            if linesIntersect(x1, y1, x2, y2, x3, y3, x4, y4):
                return True
    for x,y in line:
        if top < x < bottom or left < y < right:
            return True
    return False

def dash(x1: RealNum, y1: RealNum, x2: RealNum, y2: RealNum, d: RealNum, g: RealNum, o: RealNum=0, emptyStart: bool=False) -> List[List[Coord]]:
    """
    Finds points along a segment that are a specified distance apart.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.dash
    """
    if d <= 0 or g <= 0:
        return ValueError("dash or gap length cannot be <= 0")
    m = None if x1 == x2 else (y2-y1)/(x2-x1)
    dash = []
    gap = False

    if o == 0:
        x3, y3 = (x1, y1)
    else:
        results = pointsAway(x1, y1, o, m)
        x3, y3 = results[0] if min(x1,x2) <= results[0][0] <= max(x1,x2) and min(y1,y2) <= results[0][1] <= max(y1,y2) else results[1]
    if not emptyStart and o != 0:
        dash.append([(x1, y1), (x3, y3)])
        gap = True
    else:
        dash.append([(x3, y3)])

    while min(x1,x2) <= x3 <= max(x1,x2) and min(y1,y2) <= y3 <= max(y1,y2):
        if gap:
            results = pointsAway(x3, y3, g, m)
            if np.sign(x2-x1) == np.sign(results[0][0]-x1) and np.sign(y2-y1) == np.sign(results[0][1]-y1):
                if math.dist(results[0], (x1,y1)) > math.dist(results[1], (x1,y1)):
                    x3, y3 = results[0]
                else:
                    x3, y3 = results[1]
            else:
                x3, y3 = results[1]
            #x3, y3 = results[0] if np.sign(x2-x1) == np.sign(results[0][0]-x1) and np.sign(y2-y1) == np.sign(results[0][1]-y1) and math.dist(results[0], (x1,y1)) > math.dist(results[1], (x1,y1)) else results[1]
            dash.append([(x3, y3)])
            gap = False
        else:
            results = pointsAway(x3, y3, d, m)
            #print(results)
            #print(np.sign(x2-x1) == np.sign(results[0][0]-x1) and np.sign(y2-y1) == np.sign(results[0][1]-y1))
            if np.sign(x2-x1) == np.sign(results[0][0]-x1) and np.sign(y2-y1) == np.sign(results[0][1]-y1):
                if not (np.sign(x2-x1) == np.sign(results[1][0]-x1) and np.sign(y2-y1) == np.sign(results[1][1]-y1)) or math.dist(results[0], (x1,y1)) > math.dist(results[1], (x1,y1)):
                    x3, y3 = results[0]
                else:
                    x3, y3 = results[1]
            else:
                x3, y3 = results[1]
            #x3, y3 = results[0] if np.sign(x2-x1) == np.sign(results[0][0]-x1) and np.sign(y2-y1) == np.sign(results[0][1]-y1) and math.dist(results[0], (x1,y1)) > math.dist(results[1], (x1,y1)) else results[1]
            dash[-1].append((x3, y3))
            gap = True
    
    if len(dash[-1]) == 1: # last is gap
        dash.pop()
    else: # last is dash
        dash[-1][1] = (x2, y2)
        if dash[-1][0] == dash[-1][1]:
            dash.pop()

    return dash

def dashOffset(coords: List[Coord], d: RealNum, g: RealNum) -> Tuple[RealNum, bool]:
    """
    Calculates the offsets on each coord of a line for a smoother dashing sequence.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.dashOffset
    """
    o = 0
    offsets = [(0, False)]
    emptyStart = False
    left = None
    for c in range(len(coords)-1):
        dashes = dash(coords[c][0], coords[c][1], coords[c+1][0], coords[c+1][1], d, g, o, emptyStart)
        if dashes == []: #line has no dashes
            prev = 0 if offsets == [] or left == None else left
            remnant = math.dist(coords[c], coords[c+1])
            o = abs(g - prev - remnant)
            left = prev + remnant
            emptyStart = True
        elif dashes[0] == [coords[c], coords[c+1]]: #line is entirely dash
            prev = 0 if offsets == [] or left == None else left
            remnant = math.dist(coords[c], coords[c+1])
            o = abs(d - prev - remnant)
            left = prev + remnant
            emptyStart = False
        else:
            lastCoord = dashes[-1][1]
            if lastCoord == coords[c+1]: #line ended with a dash
                if round(remnant := math.dist(dashes[-1][0], lastCoord), 10) == d: #last dash is exactly d
                    left = 0
                    o = 0
                    emptyStart = True
                else:
                    left = remnant
                    o = d - remnant
                    emptyStart = False
            else: #line ended with a gap
                if round(remnant := math.dist(dashes[-1][0], coords[c+1]), 10) == g: #last gap is exactly g
                    left = 0
                    o = 0
                    emptyStart = False
                else:
                    o = g - remnant
                    left = remnant
                    emptyStart = True
        offsets.append((round(o, 2), emptyStart))
    return offsets

def rotateAroundPivot(x: RealNum, y: RealNum, px: RealNum, py: RealNum, theta: RealNum) -> Coord:
    """
    Rotates a set of coordinates around a pivot point.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.rotateAroundPivot
    """
    #provide θ in degrees
    theta = math.radians(theta)
    x -= px
    y -= py
    nx = x*math.cos(theta) - y*math.sin(theta)
    ny = y*math.cos(theta) + x*math.sin(theta)
    nx += px
    ny += py
    return nx, ny

def pointsAway(x: RealNum, y: RealNum, d: RealNum, m: RealNum) -> List[Coord]:
    """
    Finds two points that are a specified distance away from a specified point, all on a straight line.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.mathtools.pointsAway
    """
    theta = math.atan(m) if m != None else math.pi/2
    dx = round(d*math.cos(theta), 10)
    dy = round(d*math.sin(theta), 10)
    return [(x+dx, y+dy), (x-dx, y-dy)]


import blessed
import re
import json
import click
import difflib
term = blessed.Terminal()

import renderer.internals.internal as internal
import renderer.misc as misc

print(term.yellow("Welcome to the plaJson builder!\n-------------------------------"))

plas, plaFile = internal.askFileName("PLA") # pylint: disable=no-member
nodes, _ = internal.askFileName("Node") # pylint: disable=no-member
skinName = input(term.yellow("Name of skin [blank for default]: "))
try:
    skin = misc.getSkin(skinName if skinName != '' else 'default')
except FileNotFoundError:
    skin = misc.getSkin('default')

newPlas = {}
e = False
while not e:
    nameConfirmed = False
    while not nameConfirmed:
        name = input(term.yellow("Name of new PLA (type 'exit' to exit): "))
        if name in plas.keys() or name in newPlas.keys():
            print(term.red("PLA already exists; do you want to override its current value?"))
            print(term.red(plas[name] if name in plas.keys() else newPlas[name]))
            if input(term.red("Type 'y' to confirm: ")) != "y":
                continue
        nameConfirmed = True
    
    if name == "exit":
        e = True
        print(term.yellow("Exited"))
        continue

    typeConfirmed = False
    while not typeConfirmed:
        type_ = input(term.yellow(f"PLA type of {name}: "))
        if not type_.split(" ")[0] in skin['types'].keys():
            print(term.red(f"Type {type_.split(' ')[0]} does not exist"))
            internal.similar(type_.split(' ')[0], skin['types'].keys()) # pylint: disable=no-member
            continue
        typeConfirmed = True
    displayname = input(term.yellow(f"Display name of {name}: "))
    description = input(term.yellow(f"Description of {name}: "))
    try:
        layer = float(input(term.yellow(f"Layer for {name}: ")))
    except ValueError: 
        layer = 0
    layer = int(layer) if int(layer) == layer else layer

    if skin['types'][type_.split(" ")[0]]['type'] == "point":
        nodeConfirmed = False   
        while not nodeConfirmed:
            node = input(term.yellow("Node attached: "))
            if node not in nodes.keys():
                print(term.red("Node does not exist"))
                internal.similar(node, nodes) # pylint: disable=no-member
                continue
            nodeConfirmed = True
        newNodes = [node]
    else:
        print(term.yellow("Nodes for {name}: Choose what to pre-write in node list editor: "), end="")
        option = input(term.yellow(f"Input\n'r' for regex search of nodes\n's' for basic search of nodes\n'a' to include all nodes\n'n' for an automatic and sorted search (based on name+number)\nanything else for nothing\n"))
        if option == 'r':
            regex = input(term.yellow("Regex: "))
            preNodes = filter(lambda x: re.search(regex, x), nodes.keys())
        elif option == 's':
            substring = input(term.yellow("Substring: "))
            preNodes = filter(lambda x: substring in x, nodes.keys())
        elif option == 'a':
            preNodes = list(nodes.keys())
        elif option == 'n':
            index = []
            for n in nodes.keys():
                if ss := re.search(fr"{name}(\d*)\d", n):
                    index.append((ss.group(1) if ss.group(1) != '' else 0, n))
            preNodes = [n[1] for n in sorted(index, key=lambda a,b: a[0]-b[0])]
        else:
            preNodes = []
        print(term.yellow("Your text editor will open; input a list of nodes, one node per line. ") + term.bright_yellow("Remember to save before closing."))
        nodeConfirmed = False
        while not nodeConfirmed:
            input(term.yellow("Press enter to launch editor..."))
            nodeList = click.edit('\n'.join(preNodes), require_save=False).strip()
            if nodeList == "":
                print(term.red("No nodes given"))
                continue

            class exiter(Exception):
                pass
            try:
                for n in nodeList.split('\n'):
                    if n not in nodes.keys():
                        print(term.red(f"Node {n} does not exist"))
                        internal.similar(n, nodes.keys()) # pylint: disable=no-member
                        raise exiter
            except exiter:
                preNodes = nodeList.split('\n')
                continue
            newNodes = nodeList.split('\n')
            nodeConfirmed = True

    newPla = {
        name: {
            'type': type_,
            'displayname': displayname,
            'description': description,
            'layer': layer,
            'nodes': newNodes,
            'attrs': {}
        }
    }

    print(term.yellow("Confirm new PLA (type 'c' to cancel, anything else to confirm):"))
    res = input(term.yellow(str(newPla)))
    if res == 'c':
        print(term.yellow("Cancelled"))
        continue
    newPlas.update(newPla)

with open(plaFile, "r+") as f:
    d = json.load(f)
    d.update(newPlas)
    f.seek(0)
    f.truncate()
    json.dump(d, f, indent=4)
    f.close()

print(term.yellow("Written to " + plaFile))
    



import blessed
from schema import Schema, And, Or, Regex, Optional
from typing import Union, List, Tuple

import renderer.internals.internal as internal # type: ignore
import renderer.tools as tools
import renderer.mathtools as mathtools
import renderer.misc as misc

RealNum = Union[int, float]
Coord = Tuple[RealNum, RealNum]
TileCoord = Tuple[int, int, int]

def vCoords(coords: List[Coord]) -> True:
    """
    Validates a list of coordinates.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.validate.coords
    """
    for item in coords:
        if not isinstance(item, (tuple, list)):
            raise TypeError(f"Coordinates {item} is not type 'tuple'")
        elif len(item) != 2:
            raise ValueError(f"Coordinates {item} has {len(item)} values instead of 2")
        for n in item:
            if not isinstance(n, (int, float)):
                raise TypeError(f"Coordinate {n} is not type 'int/float'")
    return True

def vTileCoords(tiles: List[TileCoord], minZoom: int, maxZoom: int) -> True:
    """
    Validates a list of tile coordinates.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.validate.tileCoords
    """
    for item in tiles:
        if not isinstance(item, tuple):
            raise TypeError(f"Tile coordinates {item} is not type 'tuple'")
        elif len(item) != 3:
            raise ValueError(f"Tile coordinates {item} has {len(item)} values instead of 3")
        for n in item:
            if not isinstance(n, (int, float)):
                raise TypeError(f"Tile coordinate {n} is not type 'int/float'")
        if not minZoom <= item[0] <= maxZoom:
            raise ValueError(f"Zoom value {item[0]} is not in the range {minZoom} <= z <= {maxZoom}")
        elif not isinstance(item[0], int):
            raise TypeError(f"Zoom value {item[0]} is not an integer")
    return True

def vNodeList(nodes: List[str], nodeList: dict) -> True:
    """
    Validates a list of node IDs.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.validate.nodeList
    """
    for node in nodes:
        if not node in nodeList.keys():
            raise ValueError(f"Node '{node}' does not exist")

    return True

def vNodeJson(nodeList: dict) -> True:
    """
    Validates a dictionary/JSON of nodes.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.validate.nodeJson
    """
    schema = Schema({
        str: {
            "x": Or(int, float),
            "y": Or(int, float),
            "connections": list
        }
    })
    schema.validate(nodeList)
    return True

def vPlaJson(plaList: dict, nodeList: dict) -> True:
    """
    Validates a dictionary/JSON of PLAs.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.validate.plaJson
    """
    schema = Schema({
        str: {
            "type": str,
            "displayname": str,
            "description": str,
            "layer": Or(int, float),
            "nodes": And(list, lambda i: vNodeList(i, nodeList)),
            Optional("hollows"): [And(list, lambda i: vNodeList(i, nodeList))],
            "attrs": dict
        }
    })
    schema.validate(plaList)
    return True
            
def vSkinJson(skinJson: dict) -> True:
    """
    Validates a skin JSON file.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.validate.skinJson
    """
    mainSchema = Schema({
        "info": {
            "size": int,
            "font": {
                "": str,
                "b": str,
                "i": str,
                "bi": str
            },
            "background": And([int], lambda l: len(l) == 3 and not False in [0 <= n <= 255 for n in l])
        },
        "order": [str],
        "types": {
            str: {
                "tags": list,
                "type": lambda s: s in ['point', 'line', 'area'],
                "style": {
                    str: list
                }
            }
        }
    })
    point_circle = Schema({
        "layer": "circle",
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "outline": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "size": int,
        "width": int
    })
    point_text = Schema({
        "layer": "text",
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "offset": And([int], lambda o: len(o) == 2),
        "size": int,
        "anchor": Or(None, str)
    })
    point_square = Schema({
        "layer": "square",
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "outline": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "size": int,
        "width": int
    })
    point_image = Schema({
        "layer": "image",
        "file": str,
        "offset": And([int], lambda o: len(o) == 2)
    })
    line_backfore = Schema({
        "layer": Or("back", "fore"),
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "width": int,
        Optional("dash"): And([int], lambda l: len(l) == 2)
    })
    line_text = Schema({
        "layer": "text",
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "size": int,
        "offset": int
    })
    area_fill = Schema({
        "layer": "fill",
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "outline": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        Optional("stripe"): And([int], lambda l: len(l) == 3)
    })
    area_bordertext = Schema({
        "layer": "bordertext",
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "offset": int,
        "size": int
    })
    area_centertext = Schema({
        "layer": "centertext",
        "colour": Or(None, And(str, Regex(r'^#[a-f,0-9]{6}$'))),
        "size": int,
        "offset": And(And(list, [int]), lambda o: len(o) == 2)
    })
    area_centerimage = Schema({
        "layer": "image",
        "file": str,
        "offset": And(And(list, [int]), lambda o: len(o) == 2)
    })

    schemas = {
        "point": {
            "circle": point_circle,
            "text": point_text,
            "square": point_square,
            "image": point_image
        },
        "line": {
            "text": line_text,
            "back": line_backfore,
            "fore": line_backfore
        },
        "area": {
            "bordertext": area_bordertext,
            "centertext": area_centertext,
            "fill": area_fill,
            "centerimage": area_centerimage
        }
    }

    mainSchema.validate(skinJson)
    for n, t in skinJson['types'].items():
        if not n in skinJson['order']:
            raise ValueError(f"Type {n} is not in order list")
        s = t['style']
        for z, steps in s.items():
            if internal.strToTuple(z)[0] > internal.strToTuple(z)[1]:
                raise ValueError(f"Invalid range '{z}'")
            for step in steps:
                if not step["layer"] in schemas[t['type']]:
                    raise ValueError(f"Invalid layer '{step}'")
                else:
                    try:
                        schemas[t['type']][step['layer']].validate(step)
                    except Exception as e:
                        term = blessed.Terminal()
                        print(term.red(f"Type {n}, range {z}, step {step['layer']}"))
                        raise e

def vGeoJson(geoJson: dict) -> True:
    mainSchema = Schema({
        "type": "FeatureCollection", 
        "features": [{
            "type": "Feature",
            "geometry": dict,
            "properties": dict
        }]
    }, ignore_extra_keys=True)

    point = Schema({
        "type": "Point",
        "coordinates": And([int, float], lambda c: len(c)==2)
    }, ignore_extra_keys=True)

    lineString = Schema({
        "type": "LineString",
        "coordinates": vCoords
    }, ignore_extra_keys=True)

    polygon = Schema({
        "type": "Polygon",
        "coordinates": lambda cs: all([vCoords(c) and c[0] == c[-1] for c in cs])
    }, ignore_extra_keys=True)

    multiPoint = Schema({
        "type": "MultiPoint",
        "coordinates": vCoords
    }, ignore_extra_keys=True)

    multiLineString = Schema({
        "type": "MultiLineString",
        "coordinates": lambda cs: all([vCoords(c) for c in cs])
    }, ignore_extra_keys=True)

    multiPolygon = Schema({
        "type": "MultiPolygon",
        "coordinates": lambda css: all([all([vCoords(c) and c[0] == c[-1] for c in cs]) for cs in css])
    }, ignore_extra_keys=True)

    def vGeometry(geo: dict):
        schemas = {
            "Point": point,
            "LineString": lineString,
            "Polygon": polygon,
            "MultiPoint": multiPoint,
            "MultiLineString": multiLineString,
            "MultiPolygon": multiPolygon
        }

        if geo['type'] == "GeometryCollection":
            for sgeo in geo['geometries']:
                vGeometry(sgeo)
        elif geo['type'] in schemas.keys():
            schemas[geo['type']].validate(geo)
        else:
            raise ValueError(f"Invalid type {geo['type']}")

    mainSchema.validate(geoJson)
    for feature in geoJson['features']:
        vGeometry(feature['geometry'])

    return True


import os

import renderer.internals.internal as internal # type: ignore
import renderer.tools as tools
import renderer.validate as validate
import renderer.mathtools as mathtools

def getSkin(name: str) -> dict:
    """
    Gets the skin JSON, given the name.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.misc.getSkin
    """
    try:
        return internal.readJson(os.path.dirname(__file__)+"/skins/"+name+".json")
    except FileNotFoundError:
        raise FileNotFoundError(f"Skin '{name}' not found")


import argparse
import glob
import blessed
import os

def args():
    import renderer
    term = blessed.Terminal()

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(help='task to run', dest="task")

    p_info = subparsers.add_parser('info', help='view info about the renderer', formatter_class=argparse.MetavarTypeHelpFormatter)

    p_nodebuider = subparsers.add_parser('nodebuilder', help='launch the node builder', formatter_class=argparse.MetavarTypeHelpFormatter)
    p_plabuilder = subparsers.add_parser('plabuilder', help='launch the PLA builder', formatter_class=argparse.MetavarTypeHelpFormatter)

    p_validate = subparsers.add_parser('validate', help='validate a JSON file', formatter_class=argparse.MetavarTypeHelpFormatter)
    p_validate.add_argument('-p', '--pla', type=str, help='PLA JSON file to validate')
    p_validate.add_argument('-n', '--nodes', required=True, type=str, help='node JSON file to validate')

    p_vdir = subparsers.add_parser('vdir', help='validate a directory of JSON files', formatter_class=argparse.MetavarTypeHelpFormatter)
    p_vdir.add_argument('-p', '--pla', type=str, help='directory of folder of PLA JSON files to validate')
    p_vdir.add_argument('-n', '--nodes', required=True, type=str, help='directory of folder of node JSON files to validate')
    p_vdir.add_argument('-ps', '--plaSuffix', type=str, help='The suffix for PLA files\' names', default='pla')
    p_vdir.add_argument('-ns', '--nodesSuffix', type=str, help='The suffix for node files\' names', default='nodes')

    p_render = subparsers.add_parser('render', help='render tiles', formatter_class=argparse.MetavarTypeHelpFormatter)
    p_render.add_argument('-p', '--plaJson', required=True, type=str, help='the PLA Json file directory')
    p_render.add_argument('-n', '--nodeJson', required=True, type=str, help='the node Json file directory')
    p_render.add_argument('-s', '--skinJson', type=str, help='the name of the skin to use', default='default')
    p_render.add_argument('-min', '--minZoom', type=int, required=True, help="minimum zoom value")
    p_render.add_argument('-max', '--maxZoom', type=int, required=True, help="maximum zoom value")
    p_render.add_argument('-r', '--maxZoomRange', type=float, required=True, help="ange of coordinates covered by a tile in the maximum zoom")
    p_render.add_argument('-d', '--saveDir', type=str, help="the directory to save tiles in", default='')
    p_render.add_argument('-m', '--processes', type=int, help="the amount of processes to run for rendering", default=1)
    p_render.add_argument('-t', '--tiles', type=list, help="a list of tiles to render, given in tuples of (z,x,y)")
    p_render.add_argument('-o', '--offset', type=tuple, help="the offset of node coordinates, given as (x,y)", default=[0, 0])

    p_merge = subparsers.add_parser('merge', help='merge tiles', formatter_class=argparse.MetavarTypeHelpFormatter)
    p_merge.add_argument('-i', '--imageDir', type=str, help='the directory of tiles', default=os.getcwd())
    p_merge.add_argument('-s', '--saveDir', type=str, help='the directory to save the merged image to', default='')
    p_merge.add_argument('-z', '--zoom', type=int, nargs='*', help='the zoom levels to merge', default=[])

    args = parser.parse_args()

    if args.task == "info":
        print(term.yellow(f"tile-renderer v{renderer.__version__}"))
        print(term.yellow("Made by 7d for the OpenMRTMap project"))
        print("Github: https://github.com/MRT-Map/tile-renderer")
        print("PyPI: https://pypi.org/project/tile-renderer/")
        print("Docs: https://tile-renderer.readthedocs.io/en/latest/")
    elif args.task == "nodebuilder":
        import renderer.builders.node # type: ignore
    elif args.task == "plabuilder":
        import renderer.builders.pla # type: ignore
    elif args.task == "render" and __name__ == '__main__':
        renderer.render(renderer.internals.internal.readJson(args.plaJson),
                        renderer.internals.internal.readJson(args.nodeJson),
                        renderer.misc.getSkin(args.skinJson), args.minZoom,
                        args.maxZoom, args.maxZoomRange,
                        saveDir=args.saveDir,
                        processes=args.processes,
                        tiles=args.tiles,
                        offset=args.offset)
    elif args.task == "validate":
        n = renderer.internals.internal.readJson(args.nodes)
        if args.pla is not None:
            p = renderer.internals.internal.readJson(args.pla)
            renderer.validate.vPlaJson(p, n)
        else:
            renderer.validate.vNodeJson(n)
        print(term.green("Validated"))
    elif args.task == "vdir":
        if args.pla is not None:
            if not args.pla.endswith("/"): args.pla += "/"
            plas = {}
            for d in glob.glob(glob.escape(args.pla)+"*.json"):
                if d.endswith(args.plaSuffix+".json"):
                    plas.update(renderer.internals.internal.readJson(d))

        if not args.nodes.endswith("/"): args.nodes += "/"
        nodes = {}
        for d in glob.glob(glob.escape(args.nodes)+"*.json"):
            if d.endswith(args.nodesSuffix+".json"):
                nodes.update(renderer.internals.internal.readJson(d))

        if args.pla is not None:
            renderer.validate.vPlaJson(plas, nodes)
        else:
            renderer.validate.vNodeJson(nodes)
        print(term.green("Validated"))
    elif args.task == "merge":
        if not args.imageDir.endswith("/") and args.imageDir != "": args.imageDir += '/'
        renderer.tileMerge(args.imageDir, saveDir=args.saveDir, zoom=args.zoom)
    else:
        parser.print_help()

args()


import math
from PIL import Image, ImageDraw, ImageFont
import time
import blessed
import multiprocessing
import re
import itertools
term = blessed.Terminal()

import renderer.internals.internal as internal #pylint: disable=import-error
import renderer.tools as tools #pylint: disable=import-error
import renderer.validate as validate #pylint: disable=import-error
import renderer.mathtools as mathtools #pylint: disable=import-error
import renderer.misc as misc #pylint: disable=import-error

def tiles(args):
    lock, operated, start, tileCoords, tilePlas, operations, plaList, nodeList, skinJson, _, maxZoom, maxZoomRange, saveImages, saveDir, assetsDir = args # _ is minZoom
    #print(operations)
    pid = multiprocessing.current_process()._identity[0] - 1
    
    def pLog(msg):
        #print(term.move_up(processes - pid + 1), end="")
        lock.acquire()
        with term.location():
            if operated.value != 0 and operations != 0:
                print(term.green(f"{internal.percentage(operated.value, operations)}% | {internal.msToTime(internal.timeRemaining(start, operated.value, operations))} left | ") + f"{pid} | {tileCoords}: " + term.bright_black(msg), end=term.clear_eos+"\r")
            else:
                print(term.green(f"00.0% | 0.0s left | ") + f"{pid} | {tileCoords}: " + term.bright_black(msg), end=term.clear_eos+"\r")
        lock.release()
        #print(term.move_down(processes - pid + 1), end="")

    if tilePlas == [{}]:
        pLog("rendered")
        return None
    
    pLog("Initialising canvas")
    size = maxZoomRange*2**(maxZoom-internal.strToTuple(tileCoords)[0])
    im = Image.new(mode = "RGBA", size = (skinJson['info']['size'], skinJson['info']['size']), color = tuple(skinJson['info']['background']))
    img = ImageDraw.Draw(im)
    textList = []
    pointsTextList = []

    def getFont(f: str, s: int):
        if f in skinJson['info']['font'].keys():
            return ImageFont.truetype(assetsDir+skinJson['info']['font'][f], s)
        raise ValueError

    for group in tilePlas:
        info = skinJson['types'][list(group.values())[0]['type'].split(" ")[0]]
        style = []
        for zoom in info['style'].keys():
            if maxZoom-internal.strToTuple(zoom)[1] <= internal.strToTuple(tileCoords)[0] <= maxZoom-internal.strToTuple(zoom)[0]:
                style = info['style'][zoom]
                break
        for step in style:
            for plaId, pla in group.items():
                coords = [(x - internal.strToTuple(tileCoords)[1] * size, y - internal.strToTuple(tileCoords)[2] * size) for x, y in tools.nodes.toCoords(pla['nodes'], nodeList)]
                coords = [(int(skinJson['info']['size'] / size * x), int(skinJson['info']['size'] / size * y)) for x, y in coords]
                
                def point_circle():
                    img.ellipse([coords[0][0]-step['size']/2+1, coords[0][1]-step['size']/2+1, coords[0][0]+step['size']/2, coords[0][1]+step['size']/2], fill=step['colour'], outline=step['outline'], width=step['width'])

                def point_text():
                    font = getFont("", step['size'])
                    textLength = int(img.textlength(pla['displayname'], font))
                    i = Image.new('RGBA', (2*textLength,2*(step['size']+4)), (0, 0, 0, 0))
                    d = ImageDraw.Draw(i)
                    d.text((textLength, step['size']+4), pla["displayname"], fill=step['colour'], font=font, anchor="mm")
                    tw, th = i.size
                    pointsTextList.append((i, coords[0][0]+step['offset'][0], coords[0][1]+step['offset'][1], tw, th, 0))
                    # font = getFont("", step['size'])
                    # img.text((coords[0][0]+step['offset'][0], coords[0][1]+step['offset'][1]), pla['displayname'], fill=step['colour'], font=font, anchor=step['anchor'])

                def point_square():
                    img.rectangle([coords[0][0]-step['size']/2+1, coords[0][1]-step['size']/2+1, coords[0][0]+step['size']/2, coords[0][1]+step['size']/2], fill=step['colour'], outline=step['outline'], width=step['width'])

                def point_image():
                    icon = Image.open(assetsDir+step['file'])
                    im.paste(icon, (int(coords[0][0]-icon.width/2+step['offset'][0]), int(coords[0][1]-icon.height/2+step['offset'][1])), icon)

                def line_text():
                    pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Calculating text length")
                    font = getFont("", step['size'])
                    textLength = int(img.textlength(pla['displayname'], font))
                    if textLength == 0:
                        textLength = int(img.textlength("----------", font))
                    for c in range(len(coords)-1):
                        #print(coords)
                        #print(mathtools.lineInBox(coords, 0, skinJson['info']['size'], 0, skinJson['info']['size']))
                        t = math.floor(math.dist(coords[c], coords[c+1])/(4*textLength))
                        t = 1 if t == 0 else t
                        if mathtools.lineInBox(coords, 0, skinJson['info']['size'], 0, skinJson['info']['size']) and 2*textLength <= math.dist(coords[c], coords[c+1]):
                            #print(mathtools.midpoint(coords[c][0], coords[c][1], coords[c+1][0], coords[c+1][1], step['offset']))     
                            pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Generating name text")
                            for tx, ty, trot in mathtools.midpoint(coords[c][0], coords[c][1], coords[c+1][0], coords[c+1][1], step['offset'], n=t):
                                i = Image.new('RGBA', (2*textLength,2*(step['size']+4)), (0, 0, 0, 0))
                                d = ImageDraw.Draw(i)
                                d.text((textLength, step['size']+4), pla["displayname"], fill=step['colour'], font=font, anchor="mm")
                                tw, th = i.size[:]
                                i = i.rotate(trot, expand=True)
                                textList.append((i, tx, ty, tw, th, trot))
                        if "oneWay" in pla['type'].split(" ")[1:] and textLength <= math.dist(coords[c], coords[c+1]):
                            pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Generating oneway arrows")
                            getFont("b", step['size'])
                            counter = 0
                            t = math.floor(math.dist(coords[c], coords[c+1])/(4*textLength))
                            for tx, ty, _ in mathtools.midpoint(coords[c][0], coords[c][1], coords[c+1][0], coords[c+1][1], step['offset'], n=2*t+1):
                                if counter % 2 == 1:
                                    counter += 1
                                    continue
                                trot = math.degrees(math.atan2(coords[c+1][0]-coords[c][0], coords[c+1][1]-coords[c][1]))
                                i = Image.new('RGBA', (2*textLength,2*(step['size']+4)), (0, 0, 0, 0))
                                d = ImageDraw.Draw(i)
                                d.text((textLength, step['size']+4), "↓", fill=step['colour'], font=font, anchor="mm")
                                tw, th = i.size[:]
                                i = i.rotate(trot, expand=True)
                                textList.append((i, tx, ty, tw, th, trot))
                                counter += 1
                            
                def line_backfore():
                    if not "dash" in step.keys():
                        pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Drawing line")
                        img.line(coords, fill=step['colour'], width=step['width'], joint="curve")
                        if not "unroundedEnds" in info['tags']:
                            img.ellipse([coords[0][0]-step['width']/2+1, coords[0][1]-step['width']/2+1, coords[0][0]+step['width']/2, coords[0][1]+step['width']/2], fill=step['colour'])
                            img.ellipse([coords[-1][0]-step['width']/2+1, coords[-1][1]-step['width']/2+1, coords[-1][0]+step['width']/2, coords[-1][1]+step['width']/2], fill=step['colour'])
                    else:
                        offsetInfo = mathtools.dashOffset(coords, step['dash'][0], step['dash'][1])
                        #print(offsetInfo)
                        for c in range(len(coords)-1):
                            pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Drawing dashes for section {c+1} of {len(coords)}")
                            o, emptyStart = offsetInfo[c]
                            for dashCoords in mathtools.dash(coords[c][0], coords[c][1], coords[c+1][0], coords[c+1][1], step['dash'][0], step['dash'][1], o, emptyStart):
                                #print(dashCoords)
                                img.line(dashCoords, fill=step['colour'], width=step['width'])                

                def area_bordertext():
                    pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Calculating text length")
                    font = getFont("", step['size'])
                    textLength = int(img.textlength(pla['displayname'].replace('\n', ''), font))
                    for c1 in range(len(coords)):
                        c2 = c1+1 if c1 != len(coords)-1 else 0
                        if mathtools.lineInBox(coords, 0, skinJson['info']['size'], 0, skinJson['info']['size']) and 2*textLength <= math.dist(coords[c1], coords[c2]):
                            #coords[c]
                            pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Midpoints calcuated")
                            t = math.floor(math.dist(coords[c1], coords[c2])/(4*textLength))
                            t = 1 if t == 0 else t
                            allPoints = mathtools.midpoint(coords[c1][0], coords[c1][1], coords[c2][0], coords[c2][1], step['offset'], n=t, returnBoth=True)
                            for n in range(0, len(allPoints), 2):
                                pLog(f"{style.index(step)+1}/{len(style)} {plaId}: {plaId}: Generating text {n+1} of {len(allPoints)} in section {c1} of {len(coords)+1}")
                                points = [allPoints[n], allPoints[n+1]]
                                if step['offset'] < 0:
                                    tx, ty, trot = points[0] if not mathtools.pointInPoly(points[0][0], points[0][1], coords) else points[1]
                                else:
                                    #print(points[0][0], points[0][1], coords)
                                    #print(mathtools.pointInPoly(points[0][0], points[0][1], coords))
                                    tx, ty, trot = points[0] if mathtools.pointInPoly(points[0][0], points[0][1], coords) else points[1]
                                i = Image.new('RGBA', (2*textLength,2*(step['size']+4)), (0, 0, 0, 0))
                                d = ImageDraw.Draw(i)
                                d.text((textLength, step['size']+4), pla["displayname"].replace('\n', ''), fill=step['colour'], font=font, anchor="mm")
                                tw, th = i.size[:]
                                i = i.rotate(trot, expand=True)
                                textList.append((i, tx, ty, tw, th, trot))

                def area_centertext():
                    pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Calculating center")
                    cx, cy = mathtools.polyCenter(coords)
                    cx += step['offset'][0]
                    cy += step['offset'][1]
                    font = getFont("", step['size'])
                    textLength = int(min(img.textlength(x, font) for x in pla['displayname'].split('\n')))

                    left = min(c[0] for c in coords)
                    right = max(c[0] for c in coords)
                    dist = right - left
                    if textLength > dist:
                        pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Breaking up string")
                        tokens = pla['displayname'].split()
                        wss = re.findall(r"\s+", pla['displayname'])
                        text = ""
                        for token, ws in list(itertools.zip_longest(tokens, wss, fillvalue='')):
                            temptext = text[:]
                            temptext += token
                            if int(img.textlength(temptext.split('\n')[-1], font)) > dist:
                                text += '\n'+token+ws
                            else:
                                text += token+ws
                        textLength = int(max(img.textlength(x, font) for x in text.split("\n")))
                        size = int(img.textsize(text, font)[1]+4)
                    else:
                        text = pla['displayname']
                        size = step['size']+4

                    i = Image.new('RGBA', (2*textLength,2*size), (0, 0, 0, 0))
                    d = ImageDraw.Draw(i)
                    cw, ch = i.size
                    d.text((textLength, size), pla["displayname"], fill=step['colour'], font=font, anchor="mm")
                    textList.append((i, cx, cy, cw, ch, 0))

                def area_fill():
                    ai = Image.new("RGBA", (skinJson['info']['size'], skinJson['info']['size']), (0, 0, 0, 0))
                    ad = ImageDraw.Draw(ai)

                    if "stripe" in step.keys():
                        pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Generating stripes")
                        xMax, xMin, yMax, yMin = tools.line.findEnds(coords)
                        xMax += xMax-xMin
                        xMin -= yMax-yMin
                        yMax += xMax-xMin
                        yMin -= yMax-yMin
                        i = Image.new("RGBA", (skinJson['info']['size'], skinJson['info']['size']), (0, 0, 0, 0))
                        d = ImageDraw.Draw(i)
                        tlx = xMin-1
                        while tlx <= xMax:
                            d.polygon([(tlx, yMin), (tlx+step['stripe'][0], yMin), (tlx+step['stripe'][0], yMax), (tlx, yMax)], fill=step['colour'])
                            tlx += step['stripe'][0]+step['stripe'][1]
                        i = i.rotate(step['stripe'][2], center=mathtools.polyCenter(coords))
                        mi = Image.new("RGBA", (skinJson['info']['size'], skinJson['info']['size']), (0, 0, 0, 0))
                        md = ImageDraw.Draw(mi)
                        md.polygon(coords, fill=step['colour'])
                        pi = Image.new("RGBA", (skinJson['info']['size'], skinJson['info']['size']), (0, 0, 0, 0))
                        pi.paste(i, (0, 0), mi)
                        ai.paste(pi, (0, 0), pi)
                    else:
                        pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Filling area")
                        ad.polygon(coords, fill=step['colour'], outline=step['outline'])

                    if 'hollows' in pla.keys():
                        for n in pla['hollows']:
                            nCoords = [(x - internal.strToTuple(tileCoords)[1] * size, y - internal.strToTuple(tileCoords)[2] * size) for x, y in tools.nodes.toCoords(n, nodeList)]
                            nCoords = [(int(skinJson['info']['size'] / size * x), int(skinJson['info']['size'] / size * y)) for x, y in nCoords]
                            ad.polygon(nCoords, fill=(0, 0, 0, 0))
                    im.paste(ai, (0, 0), ai)

                    if step['outline'] is not None:
                        pLog(f"{style.index(step)+1}/{len(style)} {plaId}: Drawing outline")
                        exteriorOutline = coords[:]
                        exteriorOutline.append(exteriorOutline[0])
                        outlines = [exteriorOutline]
                        if 'hollows' in pla.keys():
                            for n in pla['hollows']:
                                nCoords = [(x - internal.strToTuple(tileCoords)[1] * size, y - internal.strToTuple(tileCoords)[2] * size) for x, y in tools.nodes.toCoords(n, nodeList)]
                                nCoords = [(int(skinJson['info']['size'] / size * x), int(skinJson['info']['size'] / size * y)) for x, y in nCoords]
                                nCoords.append(nCoords[0])
                                outlines.append(nCoords)
                        for oCoords in outlines:
                            img.line(oCoords, fill=step['outline'], width=2, joint="curve")
                            if not "unroundedEnds" in info['tags']:
                                img.ellipse([oCoords[0][0]-2/2+1, oCoords[0][1]-2/2+1, oCoords[0][0]+2/2, oCoords[0][1]+2/2], fill=step['outline'])

                def area_centerimage():
                    x, y = mathtools.polyCenter(coords)
                    icon = Image.open(assetsDir+step['file'])
                    im.paste(i, (x+step['offset'][0], y+step['offset'][1]), icon)

                funcs = {
                    "point": {
                        "circle": point_circle,
                        "text": point_text,
                        "square": point_square,
                        "image": point_image
                    },
                    "line": {
                        "text": line_text,
                        "back": line_backfore,
                        "fore": line_backfore
                    },
                    "area": {
                        "bordertext": area_bordertext,
                        "centertext": area_centertext,
                        "fill": area_fill,
                        "centerimage": area_centerimage
                    }
                }

                if step['layer'] not in funcs[info['type']].keys():
                    raise KeyError(f"{step['layer']} is not a valid layer")
                pLog(f"{style.index(step)+1}/{len(style)} {plaId}: ")
                funcs[info['type']][step['layer']]()

                operated.value += 1

            if info['type'] == "line" and "road" in info['tags'] and step['layer'] == "back":
                pLog("Studs: Finding connected lines")
                nodes = []
                for pla in group.values():
                    nodes += pla['nodes']
                connectedPre = [tools.nodes.findPlasAttached(x, plaList) for x in nodes]
                connected = []
                for i in connectedPre:
                    connected += i
                for conPla, index in connected:
                    if not "road" in skinJson['types'][plaList[conPla]['type'].split(" ")[0]]['tags']:
                        continue
                    conInfo = skinJson['types'][plaList[conPla]['type'].split(" ")[0]]
                    conStyle = []
                    for zoom in conInfo['style'].keys():
                        if maxZoom-internal.strToTuple(zoom)[1] <= internal.strToTuple(tileCoords)[0] <= maxZoom-internal.strToTuple(zoom)[0]:
                            conStyle = conInfo['style'][zoom]
                            break
                    for conStep in conStyle:
                        if conStep['layer'] in ["back", "text"]:
                            continue
                        
                        pLog("Studs: Extracting coords")
                        conCoords = [(x-internal.strToTuple(tileCoords)[1]*size, y-internal.strToTuple(tileCoords)[2]*size) for x,y in tools.nodes.toCoords(plaList[conPla]['nodes'], nodeList)]
                        conCoords = [(int(skinJson['info']['size']/size*x), int(skinJson['info']['size']/size*y)) for x,y in conCoords]
                        preConCoords = conCoords[:]
                        
                        pLog("Studs: Coords processed")
                        if index == 0:
                            conCoords = [conCoords[0], conCoords[1]]
                            if not "dash" in conStep.keys():
                                conCoords[1] = ((conCoords[0][0]+conCoords[1][0])/2, (conCoords[0][1]+conCoords[1][1])/2)
                        elif index == len(conCoords)-1:
                            conCoords = [conCoords[index-1], conCoords[index]]
                            if not "dash" in conStep.keys():
                                conCoords[0] = ((conCoords[0][0]+conCoords[1][0])/2, (conCoords[0][1]+conCoords[1][1])/2)
                        else:
                            conCoords = [conCoords[index-1], conCoords[index], conCoords[index+1]]
                            if not "dash" in conStep.keys():
                                conCoords[0] = ((conCoords[0][0]+conCoords[1][0])/2, (conCoords[0][1]+conCoords[1][1])/2)
                                conCoords[2] = ((conCoords[2][0]+conCoords[1][0])/2, (conCoords[2][1]+conCoords[1][1])/2)
                        pLog("Studs: Segment drawn")
                        if not "dash" in conStep.keys():
                            img.line(conCoords, fill=conStep['colour'], width=conStep['width'], joint="curve")
                            img.ellipse([conCoords[0][0]-conStep['width']/2+1, conCoords[0][1]-conStep['width']/2+1, conCoords[0][0]+conStep['width']/2, conCoords[0][1]+conStep['width']/2], fill=conStep['colour'])
                            img.ellipse([conCoords[-1][0]-conStep['width']/2+1, conCoords[-1][1]-conStep['width']/2+1, conCoords[-1][0]+conStep['width']/2, conCoords[-1][1]+conStep['width']/2], fill=conStep['colour'])

                        else:
                            offsetInfo = mathtools.dashOffset(preConCoords, conStep['dash'][0], conStep['dash'][1])[index:]
                            #print(offsetInfo)
                            for c in range(len(conCoords)-1):
                                #print(offsetInfo)
                                #print(c)
                                o, emptyStart = offsetInfo[c]
                                for dashCoords in mathtools.dash(conCoords[c][0], conCoords[c][1], conCoords[c+1][0], conCoords[c+1][1], conStep['dash'][0], conStep['dash'][1], o, emptyStart):
                                    #print(dashCoords)
                                    img.line(dashCoords, fill=conStep['colour'], width=conStep['width'])
                operated.value += 1

    textList += pointsTextList 
    textList.reverse()
    dontCross = []
    processed = 0
    #print(textList)
    for i, x, y, w, h, rot in textList:
        r = lambda a,b : mathtools.rotateAroundPivot(a, b, x, y, rot)
        currentBoxCoords = [r(x-w/2, y-h/2), r(x-w/2, y+h/2), r(x+w/2, y+h/2), r(x+w/2, y-h/2), r(x-w/2, y-h/2)]
        canPrint = True
        for box in dontCross:
            _, ox, oy, ow, oh, _ = textList[dontCross.index(box)]
            oMaxDist = ((ow/2)**2+(oh/2)**2)**0.5/2
            thisMaxDist = ((w/2)**2+(h/2)**2)**0.5/2
            dist = ((x-ox)**2+(y-oy)**2)**0.5
            if dist > oMaxDist + thisMaxDist:
                continue
            for c in range(len(box)-1):
                for d in range(len(currentBoxCoords)-1):
                    canPrint = False if mathtools.linesIntersect(box[c][0], box[c][1], box[c+1][0], box[c+1][1], currentBoxCoords[d][0], currentBoxCoords[d][1], currentBoxCoords[d+1][0], currentBoxCoords[d+1][1]) else canPrint
                    if not canPrint:
                        break
                if not canPrint:
                    break
            if canPrint and mathtools.pointInPoly(currentBoxCoords[0][0], currentBoxCoords[0][1], box) or mathtools.pointInPoly(box[0][0], box[0][1], currentBoxCoords):
                canPrint = False
            if canPrint == False:
                break
        processed += 1
        if canPrint:
            pLog(f"Text {processed}/{len(textList)} pasted")
            im.paste(i, (int(x-i.width/2), int(y-i.height/2)), i)
        else:
            pLog(f"Text {processed}/{len(textList)} skipped")
        dontCross.append(currentBoxCoords)
    operated.value += 1
    
    #tileReturn[tileCoords] = im
    if saveImages:
        im.save(f'{saveDir}{tileCoords}.png', 'PNG')

    pLog("Rendered")

    return {tileCoords: im}


import math
import blessed
from typing import Union, List, Dict, Tuple
term = blessed.Terminal()

import renderer.internals.internal as internal # type: ignore
import renderer.validate as validate
import renderer.mathtools as mathtools
import renderer.misc as misc

RealNum = Union[int, float]
Coord = Tuple[RealNum, RealNum]
TileCoord = Tuple[int, int, int]

class plaJson:
    @staticmethod
    def findEnds(plaList: dict, nodeList: dict) -> Tuple[RealNum, RealNum, RealNum, RealNum]:
        """
        Finds the minimum and maximum X and Y values of a JSON or dictionary of PLAs
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.plaJson.findEnds
        """
        validate.vPlaJson(plaList, nodeList)
        validate.vNodeJson(nodeList)
        xMax = -math.inf
        xMin = math.inf
        yMax = -math.inf
        yMin = math.inf
        for pla in plaList.keys():
            coords = nodes.toCoords(plaList[pla]['nodes'], nodeList)
            for x, y in coords:
                xMax = x if x > xMax else xMax
                xMin = x if x < xMin else xMin
                yMax = y if y > yMax else yMax
                yMin = y if y < yMin else yMin
        return xMax, xMin, yMax, yMin

    @staticmethod
    def renderedIn(plaList: dict, nodeList: dict, minZoom: int, maxZoom: int, maxZoomRange: RealNum) -> List[TileCoord]:
        """
        Like renderer.tools.lineToTiles(), but for a JSON or dictionary of PLAs.
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.plaJson.renderedIn
        """
        validate.vPlaJson(plaList, nodeList)
        validate.vNodeJson(nodeList)
        if maxZoom < minZoom:
            raise ValueError("Max zoom value is lesser than min zoom value")

        tiles = []
        for pla in plaList.keys():
            coords = nodes.toCoords(plaList[pla]['nodes'], nodeList)
            tiles.extend(line.toTiles(coords, minZoom, maxZoom, maxZoomRange))

        return tiles

    @staticmethod
    def toGeoJson(plaList: dict, nodeList: dict, skinJson: dict) -> dict:
        """
        Converts PLA Json into GeoJson (with nodes and skin).
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.plaJson.toGeoJson
        """
        validate.vPlaJson(plaList, nodeList)
        validate.vNodeJson(nodeList)
        validate.vSkinJson(skinJson)
        geoJson = {"type": "FeatureCollection", "features":[]}
        
        for plaId, pla in plaList.items():
            geoFeature = {"type": "Feature"}
            if not pla['type'].split()[0] in skinJson['types'].keys():
                plaType = "UNKNOWN"
                plaShape = "UNKNOWN"
                print(term.yellow(f"WARNING: Type {pla['type']} not in skin file"))
            else:
                plaType = pla['type'].split()[0]
                plaShape = skinJson['types'][plaType]['type']
            
            if plaShape == "point":
                geoCoords = list(nodes.toCoords(pla['nodes'], nodeList)[0])
                geoShape = "Point"
            elif plaShape == "area":
                geoCoords = []
                geoCoords.append([list(c) for c in nodes.toCoords(pla['nodes'], nodeList)])
                if geoCoords[0][0] != geoCoords[0][-1]:
                    geoCoords[0].append(geoCoords[0][0])
                if 'hollows' in pla.keys():
                    for hollow in pla['hollows']:
                        geoCoords.append([list(c) for c in nodes.toCoords(hollow, nodeList)])
                        if geoCoords[-1][0] != geoCoords[-1][-1]:
                            geoCoords[-1].append(geoCoords[-1][0])
                geoShape = "Polygon"
            else:
                geoCoords = [list(c) for c in nodes.toCoords(pla['nodes'], nodeList)]
                geoShape = "LineString"
            geoFeature['geometry'] = {
                "type": geoShape,
                "coordinates": geoCoords
            }

            geoFeature['properties'] = {
                "name": plaId,
                "plaType": pla['type'],
                "displayname": pla['displayname'],
                "description": pla['description'],
                "layer": pla['layer']
            }
            for k, v in pla['attrs'].items():
                geoFeature['properties'][k] = v
        
            geoJson['features'].append(geoFeature)

        return geoJson

class geoJson:
    @staticmethod
    def toNodePlaJson(geoJson: dict) -> Tuple[dict, dict]:
        """
        Converts GeoJson to PLA and Node JSONs.
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.geoJson.toNodePlaJson
        """
        validate.vGeoJson(geoJson)
        plaJson = {}
        nodeJson = {}
        
        def addNode(x, y):
            for k, v in nodeJson.items():
                if v['x'] == x and v['y'] == y:
                    return k
            k = internal.genId()
            nodeJson[k] = {
                "x": x,
                "y": y,
                "connections": []
            }
            return k

        def singleGeometry(geo, properties, name=None):
            if name == None:
                if 'name' in properties.keys(): name = properties['name']
                else: name = internal.genId()
            plaType = properties['plaType'] if "plaType" in properties else "UNKNOWN"
            displayname = properties['displayname'] if "displayname" in properties else ""
            description = properties['description'] if "description" in properties else ""
            layer = properties['layer'] if "layer" in properties else 0
            attrs = {}
            for k, v in properties.items():
                if not k in ['name', 'plaType', 'displayname', 'description', 'layer']:
                    attrs[k] = v

            hollows = []
            if geo['type'] == "Polygon":
                nodes = [addNode(*c) for c in geo['coordinates'][0]]
                if len(geo['coordinates']) > 1:
                    for i in range(1, len(geo['coordinates'])):
                        hollows.append([addNode(*c) for c in geo['coordinates'][i]])
            elif geo['type'] == "LineString":
                nodes = [addNode(*c) for c in geo['coordinates']]
            else:
                nodes = addNode(*geo['coordinates'])
            plaJson[name] = {
                "type": plaType,
                "displayname": displayname,
                "description": description,
                "layer": layer,
                "nodes": nodes,
                "attrs": attrs
            }
            if hollows != []: plaJson[name]['hollows'] = hollows

        def singleFeature(feature: dict):
            name = feature['properties']['name'] if 'name' in feature['properties'].keys() else internal.genId()
            if feature['geometry']['type'] == "GeometryCollection":
                for itemNo, sgeo in enumerate(feature['geometry']['geometries']):
                    singleGeometry(sgeo, feature['properties'], name=name+"_"+itemNo)
            elif feature['geometry']['type'].startswith("Multi"):
                for scoord in feature['geometry']['coordinates']:
                    singleGeometry({"type": feature['geometry']['type'].replace("Multi", ""), "coordinates": scoord}, feature['properties'], name=name)
            else:
                singleGeometry(feature['geometry'], feature['properties'])

        for feature in geoJson['features']:
            singleFeature(feature)
        return plaJson, nodeJson
            
class tile:
    @staticmethod
    def findEnds(coords: List[TileCoord]) -> Tuple[RealNum, RealNum, RealNum, RealNum]:
        """
        Find the minimum and maximum x/y values of a set of tiles coords.
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.tile.findEnds
        """
        validate.vTileCoords(coords, -math.inf, math.inf)
        xMax = -math.inf
        xMin = math.inf
        yMax = -math.inf
        yMin = math.inf
        for _, x, y in coords:
            xMax = x if x > xMax else xMax
            xMin = x if x < xMin else xMin
            yMax = y if y > yMax else yMax
            yMin = y if y < yMin else yMin
        return xMax, xMin, yMax, yMin

class line:
    @staticmethod
    def findEnds(coords: List[Coord]) -> Tuple[RealNum, RealNum, RealNum, RealNum]:
        """
        Find the minimum and maximum x/y values of a set of coords.
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.line.findEnds
        """
        validate.vCoords(coords)
        xMax = -math.inf
        xMin = math.inf
        yMax = -math.inf
        yMin = math.inf
        for x, y in coords:
            xMax = x if x > xMax else xMax
            xMin = x if x < xMin else xMin
            yMax = y if y > yMax else yMax
            yMin = y if y < yMin else yMin
        return xMax, xMin, yMax, yMin

    @staticmethod
    def toTiles(coords: List[Coord], minZoom: int, maxZoom: int, maxZoomRange: RealNum) -> List[TileCoord]:
        """
        Generates tile coordinates from list of regular coordinates using renderer.tools.coordToTiles().
        More info: Mainly for rendering whole PLAs. https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.line.toTiles
        """
        validate.vCoords(coords)
        if len(coords) == 0:
            raise ValueError("Empty list of coords given")
        elif maxZoom < minZoom:
            raise ValueError("Max zoom value is lesser than min zoom value")
        
        tiles = []
        xMax = -math.inf
        xMin = math.inf
        yMax = -math.inf
        yMin = math.inf
        
        for x, y in coords:
            xMax = x+10 if x > xMax else xMax
            xMin = x-10 if x < xMin else xMin
            yMax = y+10 if y > yMax else yMax
            yMin = y-10 if y < yMin else yMin
        xr = list(range(xMin, xMax+1, int(maxZoomRange/2)))
        xr.append(xMax+1)
        yr = list(range(yMin, yMax+1, int(maxZoomRange/2)))
        yr.append(yMax+1)
        for x in xr:
            for y in yr:
                tiles.extend(coord.toTiles([x,y], minZoom, maxZoom, maxZoomRange))
        tiles = list(dict.fromkeys(tiles))
        return tiles

class nodes:
    @staticmethod
    def findPlasAttached(nodeId: str, plaList: dict) -> List[Tuple[str, int]]:
        """
        Finds which PLAs attach to a node.
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.nodes.findPlasAttached
        """
        plas = []
        for plaId, pla in plaList.items():
            #print(plaId)
            if nodeId in pla['nodes']:
                plas.append((plaId, pla['nodes'].index(nodeId)))
        return plas

    @staticmethod
    def toCoords(nodes: List[str], nodeList: dict) -> List[Coord]:
        """
        Converts a list of nodes IDs into a list of coordinates with a node dictionary/JSON as its reference.
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.nodes.toCoords
        """
        validate.vNodeJson(nodeList)
        validate.vNodeList(nodes, nodeList)
        coords = []
        for nodeid in nodes:
            if not nodeid in nodeList.keys():
                raise KeyError(f"Node '{nodeid}' does not exist")
            coords.append((nodeList[nodeid]['x'],nodeList[nodeid]['y']))
        return coords

class coord:
    @staticmethod
    def toTiles(coord: Coord, minZoom: int, maxZoom: int, maxZoomRange: RealNum) -> List[TileCoord]:
        """
        Returns all tiles in the form of tile coordinates that contain the provided regular coordinate.
        More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.tools.coord.toTiles
        """
        if maxZoom < minZoom:
            raise ValueError("Max zoom value is lesser than min zoom value")

        tiles = []
        for z in reversed(range(minZoom, maxZoom+1)):
            x = math.floor(coord[0] / maxZoomRange)
            y = math.floor(coord[1] / maxZoomRange)
            tiles.append((z,x,y))
            maxZoomRange *= 2

        return tiles


from typing import Union, List, Dict, Tuple, Optional
from PIL import Image
import time
import glob
import re
import multiprocessing
import sys
import blessed
import os

import renderer.internals.internal as internal # type: ignore
import renderer.tools as tools
import renderer.validate as validate
import renderer.mathtools as mathtools
import renderer.internals.rendering as rendering # type: ignore
import renderer.misc as misc

RealNum = Union[int, float]
Coord = Tuple[RealNum, RealNum]
TileCoord = Tuple[int, int, int]

def tileMerge(images: Union[str, Dict[str, Image.Image]], saveImages: bool=True, saveDir: str="tiles/", zoom: List[int]=[]) -> List[Image.Image]:
    """
    Merges tiles rendered by renderer.render().
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#tileMerge
    """
    term = blessed.Terminal()
    imageDict = {}
    tileReturn = {}
    if isinstance(images, str):
        print(term.green("Retrieving images..."), end="\r")
        for d in glob.glob(glob.escape(images)+"*.png"):
            regex = re.search(r"(-?\d+, -?\d+, -?\d+)\.png$", d) # pylint: disable=anomalous-backslash-in-string
            if regex == None:
                continue
            coord = regex.group(1)
            i = Image.open(d)
            imageDict[coord] = i
            with term.location(): print(term.green("Retrieving images... ") + f"Retrieved {coord}", end=term.clear_eos+"\r")
    else:
        imageDict = images
    print(term.green("\nAll images retrieved"))
    print(term.green("Determined zoom levels..."), end=" ")
    if zoom == []:
        for c in imageDict.keys():
            z = internal.strToTuple(c)[0]
            if not z in zoom:
                zoom.append(z)
    print(term.green("determined"))
    for z in zoom:
        print(term.green(f"Zoom {z}: ") + "Determining tiles to be merged", end=term.clear_eos+"\r")
        toMerge = {}
        for c, i in imageDict.items():
            #i = imageDict[c]
            if internal.strToTuple(c)[0] == z:
                toMerge[c] = i

        tileCoords = [internal.strToTuple(s) for s in toMerge.keys()]
        #print(tileCoords)
        xMax, xMin, yMax, yMin = tools.tile.findEnds(tileCoords)
        #print(imageDict.values())
        tileSize = list(imageDict.values())[0].size[0]
        i = Image.new('RGBA', (tileSize*(xMax-xMin+1), tileSize*(yMax-yMin+1)), (0, 0, 0, 0))
        px = 0
        py = 0
        merged = 0
        start = time.time() * 1000
        for x in range(xMin, xMax+1):
            for y in range(yMin, yMax+1):
                if f"{z}, {x}, {y}" in toMerge.keys():
                    i.paste(toMerge[f"{z}, {x}, {y}"], (px, py))
                    merged += 1
                    with term.location(): print(term.green(f"Zoom {z}: ") + f"{internal.percentage(merged, len(toMerge.keys()))}% | {internal.msToTime(internal.timeRemaining(start, merged, len(toMerge.keys())))} left | " + term.bright_black(f"Pasted {z}, {x}, {y}"), end=term.clear_eos+"\r")
                py += tileSize
            px += tileSize
            py = 0
        #tileReturn[tilePlas] = im
        if saveImages:
            print(term.green(f"Zoom {z}: ") + "Saving image", end=term.clear_eos+"\r")
            i.save(f'{saveDir}merge_{z}.png', 'PNG')
        tileReturn[str(z)] = i
        
    print(term.green("\nAll merges complete"))
    return tileReturn

def render(plaList: dict, nodeList: dict, skinJson: dict, minZoom: int, maxZoom: int, maxZoomRange: RealNum, saveImages: bool=True, saveDir: str="", assetsDir: str=os.path.dirname(__file__)+"/skins/assets/", \
    processes: int=1, tiles: Optional[List[TileCoord]]=None, offset: Tuple[RealNum, RealNum]=(0,0)) -> Dict[str, Image.Image]:
    """
    Renders tiles from given coordinates and zoom values.
    More info: https://tile-renderer.readthedocs.io/en/latest/functions.html#renderer.render
    """
    if maxZoom < minZoom:
        raise ValueError("Max zoom value is greater than min zoom value")
    term = blessed.Terminal()

    # validation
    print(term.green("Validating skin..."), end=" ")
    validate.vSkinJson(skinJson)
    print(term.green("validated\nValidating nodes..."), end=" ")
    validate.vNodeJson(nodeList)
    print(term.green("validated\nValidating PLAs..."), end=" ")
    validate.vPlaJson(plaList, nodeList)
    

    # offset
    for node in nodeList.keys():
        nodeList[node]['x'] += offset[0]
        nodeList[node]['y'] += offset[1]

    print(term.green("validated\nFinding tiles..."), end=" ")
    #finds which tiles to render
    if tiles != None:
        validate.vTileCoords(tiles, minZoom, maxZoom)
    else: #finds box of tiles
        tiles = tools.plaJson.renderedIn(plaList, nodeList, minZoom, maxZoom, maxZoomRange)

    print(term.green("found\nRemoving PLAs with unknown type..."), end=" ")
    # remove PLAs whose type is not in skin
    removeList = []
    for pla in plaList.keys():
        if not plaList[pla]['type'].split(' ')[0] in skinJson['order']:
            removeList.append(pla)
    for pla in removeList:
        del plaList[pla]
    print(term.green("removed"))
    if removeList != []:
        print(term.yellow('The following PLAs were removed:'))
        print(term.yellow(" | ".join(removeList)))

    print(term.green("Sorting PLA by tiles..."), end=" ")
    #sort PLAs by tiles
    tileList = {}
    for tile in tiles:
        tileList[internal.tupleToStr(tile)] = {}
    for pla in plaList.keys():
        coords = tools.nodes.toCoords(plaList[pla]['nodes'], nodeList)
        renderedIn = tools.line.toTiles(coords, minZoom, maxZoom, maxZoomRange)
        for tile in renderedIn:
            if internal.tupleToStr(tile) in tileList.keys():
                tileList[internal.tupleToStr(tile)][pla] = plaList[pla]
    
    processStart = time.time() * 1000
    processed = 0
    timeLeft = 0.0
    l = lambda processed, timeLeft, tilePlas, msg: term.green(f"{internal.percentage(processed, len(tileList))}% | {internal.msToTime(timeLeft)} left | ") + f"{tilePlas}: " + term.bright_black(msg) + term.clear_eos
    print(term.green("sorted\n")+term.bright_green("Starting processing"))
    for tilePlas in tileList.keys():
        #sort PLAs in tiles by layer
        newTilePlas = {}
        for pla in tileList[tilePlas].keys():
            if not str(float(tileList[tilePlas][pla]['layer'])) in newTilePlas.keys():
                newTilePlas[str(float(tileList[tilePlas][pla]['layer']))] = {}
            newTilePlas[str(float(tileList[tilePlas][pla]['layer']))][pla] = tileList[tilePlas][pla]
        with term.location(): print(l(processed, timeLeft, tilePlas, "Sorted PLA by layer"), end="\r")
        

        #sort PLAs in layers in files by type
        for layer in newTilePlas.keys():
            #print(newTilePlas[layer].items())
            newTilePlas[layer] = {k: v for k, v in sorted(newTilePlas[layer].items(), key=lambda x: skinJson['order'].index(x[1]['type'].split(' ')[0]))}
        with term.location(): print(l(processed, timeLeft, tilePlas, "Sorted PLA by type"), end="\r")
        
        #merge layers
        tileList[tilePlas] = {}
        layers = sorted(newTilePlas.keys(), key=lambda x: float(x))
        for layer in layers:
            for key, pla in newTilePlas[layer].items():
                tileList[tilePlas][key] = pla
        with term.location(): print(l(processed, timeLeft, tilePlas, "Merged layers"), end="\r")

        #groups PLAs of the same type if "road" tag present
        newerTilePlas = [{}]
        keys = list(tileList[tilePlas].keys())
        for i in range(len(tileList[tilePlas])):
            newerTilePlas[-1][keys[i]] = tileList[tilePlas][keys[i]]
            if i != len(keys)-1 and (tileList[tilePlas][keys[i+1]]['type'].split(' ')[0] != tileList[tilePlas][keys[i]]['type'].split(' ')[0] or not "road" in skinJson['types'][tileList[tilePlas][keys[i]]['type'].split(' ')[0]]['tags']):
                newerTilePlas.append({})
        tileList[tilePlas] = newerTilePlas
        with term.location(): print(l(processed, timeLeft, tilePlas, "PLAs grouped"), end="\r")

        processed += 1
        timeLeft = internal.timeRemaining(processStart, processed, len(tileList))
        #timeLeft = round(((int(round(time.time() * 1000)) - processStart) / processed * (len(tileList) - processed)), 2)
    
    print(term.green("100.00% | 0.0s left | ") + "Processing complete" + term.clear_eos)

    #count # of rendering operations
    print(term.bright_green("Counting no. of operations"))
    operations = 0
    opTiles = {}
    tileOperations = 0
    for tilePlas in tileList.keys():
        if tileList[tilePlas] == [{}]:
            opTiles[tilePlas] = 0
            continue

        for group in tileList[tilePlas]:
            info = skinJson['types'][list(group.values())[0]['type'].split(" ")[0]]
            style = []
            for zoom in info['style'].keys():
                if maxZoom-internal.strToTuple(zoom)[1] <= internal.strToTuple(tilePlas)[0] <= maxZoom-internal.strToTuple(zoom)[0]:
                    style = info['style'][zoom]
                    break
            for step in style:
                for _, pla in group.items():
                    operations += 1; tileOperations += 1
                if info['type'] == "line" and "road" in info['tags'] and step['layer'] == "back":
                    operations += 1; tileOperations += 1
        operations += 1; tileOperations += 1 #text

        opTiles[tilePlas] = tileOperations
        tileOperations = 0
        print(f"Counted {tilePlas}", end=term.clear_eos+"\r")
    
    #render
    renderStart = time.time() * 1000
    print(term.bright_green("\nStarting render"))
    if __name__ == 'renderer.base':
        m = multiprocessing.Manager()
        operated = m.Value('i', 0)
        lock = m.Lock() # pylint: disable=no-member

        input = []
        for i in tileList.keys():
            input.append((lock, operated, renderStart, i, tileList[i], operations, plaList, nodeList, skinJson, minZoom, maxZoom, maxZoomRange, saveImages, saveDir, assetsDir))
        p = multiprocessing.Pool(processes)
        try:
            preresult = p.map(rendering.tiles, input)
        except KeyboardInterrupt:
            p.terminate()
            sys.exit()
        result = {}
        for i in preresult:
            if i == None:
                continue
            k, v = list(i.items())[0]
            result[k] = v

        print(term.green("100.00% | 0.0s left | ") + "Rendering complete" + term.clear_eos)
    print(term.bright_green("Render complete"))
    return result


import math
import json
from typing import Union
import time
import random
import os
import blessed
import difflib
term = blessed.Terminal()

def dictIndex(d: dict, v):
    return list(d.keys())[list(d.values()).index(v)]

def readJson(file: str):
    with open(file, "r") as f:
        data = json.load(f)
        f.close()
        return data

def writeJson(file: str, data: dict, pp=False):
    with open(file, "r+") as f:
        f.seek(0)
        f.truncate()
        if pp:
            json.dump(data, f, indent=2)
        else:
            json.dump(data, f)
        f.close()

def tupleToStr(t: tuple):
    return str(t)[1:-1]

def strToTuple(s: str):
    return tuple([int(x) for x in s.split(", ")])

def msToTime(ms: Union[int, float]):
    if ms == 0:
        return "00.0s"
    s = round(ms / 1000, 1)
    #ms = round(ms % 1000, 2)
    m = math.floor(s / 60)
    s = s % 60
    h = math.floor(m / 60)
    m = m % 60
    d = math.floor(h / 24)
    h = h % 24
    res = ""
    if d != 0:
        res = res + str(d) + "d "
    if h != 0:
        zero = "0" if h < 10 else ""
        res = res + zero + str(h) + "h "
    if m != 0:
        zero = "0" if m < 10 else ""
        res = res + zero + str(m) + "min "
    if s != 0:
        zero = "0" if s < 10 else ""
        res = res + zero + str(round(s, 1)) + "s "
    if res == "":
        res = "00.0s"
    #if ms != 0:
    #    pzero = "00" if ms < 10 else "0" if 10 <= ms < 100 else ""
    #    szero = "0" if len(str(ms).split(".")[1]) == 1 else ""
    #    res = res + pzero + str(ms) + szero + "ms "
    return res.strip()

def percentage(c: Union[int, float], t: Union[int, float]):
    res = round(c/t*100, 2)
    pzero = "0" if res < 10 else ""
    szero = "0" if len(str(res).split(".")[1]) == 1 else ""
    return pzero + str(res) + szero

def timeRemaining(start: Union[int, float], c: Union[int, float], t: Union[int, float]):
    return round(((int(round(time.time() * 1000)) - start) / c * (t - c)), 2)

def genId():
    def b10_b64(n: int):
        BASE64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
        q = n
        o = ""
        while True:
            o += BASE64[q % 64]
            q = math.floor(q / 64)
            if q == 0:
                break
        return o[::-1]
    decimalId = int(time.time() * 10000000)
    return b10_b64(decimalId) + "-" + b10_b64(random.randint(1, 64**5))

def askFileName(name: str):
    fileConfirmed = False
    while not fileConfirmed:
        filePath = input(term.yellow(f"Which {name} JSON file are you writing to/referencing? "))
        try:
            open(filePath, "r")
            if filePath.endswith(".json"):
                fileConfirmed = True
            else:
                print(term.red("File is not a JSON file"))
        except FileNotFoundError:
            print(term.red("File does not exist"))
    
    with open(filePath, "r") as f:
        data = json.load(f)
        f.close()

    return data, filePath

def similar(s, i):
    if len(sim := difflib.get_close_matches(s, i)) != 0:
        print(term.bright_red(f"Perhaps you mean: {', '.join(sim)}"))


import blessed
import re
import json
term = blessed.Terminal()

import renderer.internals.internal as internal

print(term.yellow("Welcome to the nodeJson builder!\n--------------------------------"))

nodes, filePath = internal.askFileName("Node") # pylint: disable=no-member

print(term.yellow("Ingame, press F3+C once, and paste it here.\nType 'exit' to exit."))
newNodes = {}
e = False
while not e:
    pasted = input(term.yellow("Paste: "))
    if pasted == "exit":
        e = True
        print(term.yellow("Exited"))
        continue
    groups = re.search(r"@s (\S+) \S+ (\S+)", pasted)
    if groups == None:
        print(term.red("Invalid paste"))
        continue
    x = int(float(groups.group(1)))
    y = int(float(groups.group(2)))

    nameConfirmed = False
    while not nameConfirmed:
        name = input(term.yellow("Node name: "))
        if name in nodes.keys() or name in newNodes.keys():
            print(term.red("Node already exists; do you want to override its current value?"))
            print(term.red(nodes[name] if name in nodes.keys() else newNodes[name]))
            if input(term.red("Type 'y' to confirm: ")) != "y":
                continue
        nameConfirmed = True

    newNodes[name] = {"x": x, "y": y, "connections": []}

with open(filePath, "r+") as f:
    d = json.load(f)
    d.update(newNodes)
    f.seek(0)
    f.truncate()
    json.dump(d, f, indent=4)
    f.close()

print(term.yellow("Written to " + filePath))


from setuptools import setup
import renderer

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
  name = 'tile-renderer',
  packages = ['renderer'],
  version = renderer.__version__+".1",
  license ='gpl-3.0',
  description = 'Leaflet.js streetmap tile renderer',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = '7d (from Minecart Rapid Transit Mapping Team)',
  author_email = 'i.third.7d@protonmail.com',
  url = 'https://github.com/MRT-Map/tile-renderer',
  download_url = f'https://github.com/MRT-Map/tile-renderer/archive/refs/tags/v{renderer.__version__}.tar.gz',
  keywords = ['leaflet', 'leaflet.js', 'leafletjs', 'map', 'tiles', 'renderer', 'tile-renderer', 'mapping'],
  python_requires='>=3.8',
  package_data={
    'renderer': ['skins/*', 'skins/assets/*', 'internals/*', 'builders/*'],
  },
  install_requires=[
    'pillow',
    'blessed',
    'sympy',
    'schema',
    'numpy',
    'click'
  ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
  ],
  entry_points = {
    'console_scripts': ['renderer=renderer.__main__:args'],
  }
)